/*
 * Part of the solution for Assignment 3 (CSC 218, Winter 2003), by
 * Stefan Bruda.
 */
#include "btree.h"

/***
 *** Public member functions:
 ***/

btree::btree(void) {
    the_tree = empty;
}

// The copy constructor is just a wrapped for clone.
btree::btree(const btree& tree) {
    // No need to zap the_tree, since we are inside a constructor.
    // Actually, zapping here is incorrect, because the_tree is not
    // necessarily NULL--it is merely uninitialized, which means it
    // contains an arbitrary value.  Deallocating the content of this
    // value leads more often than not to a segfault.
    the_tree = clone(tree.the_tree,empty);
}

// The copy destructor is just a wrapper for the good ol' deltree.
btree::~btree(void) {
    the_tree = deltree(the_tree,the_tree);
}

// Again a wrapper for clone.
const btree& btree::operator=(const btree& rhs) {
    if (this != &rhs) { // standard alias test
        // get rid of old value:
        the_tree = deltree(the_tree,the_tree);
        // the actual copying process:
        the_tree = clone(rhs.the_tree, empty);
    }
    return *this;  // for chained assignments
}

// Again, a wrapper for (the goo ol') deltree.
void btree::zap(btree::node* node) {
    the_tree = deltree(the_tree, node);
}

// Wrappers for printtree (which is slightly modified from the
// previous assignment).
void btree::print(ostream& out, int deflated) const {
    if (deflated)
        traverse_inorder(the_tree, out);
    else
        printtree(the_tree, out);
}

ostream& operator<< (ostream& out, const btree& tree) {
    tree.print(out);
    return out;
}

/***
 *** New helper functions: 
 ***/

void btree::traverse_inorder(const btree::node* what, ostream& out) const {
    if (what) {
        traverse_inorder(lefttree(what));
        out << key(what) << "\n";
        traverse_inorder(righttree(what));
    }
}

// Does the effective deep copy.
btree::node* btree::clone(btree::node* node, btree::node* parent) const {
    if (node == empty)
        return empty;
    btree::node* ret = mknode(node -> info); // cannot do a direct recursive call here, 
                                            // since we do not have the parent yet!
    btree::node* left = clone(node -> left, ret);
    btree::node* right = clone(node -> right, ret);
    ret -> left = left;
    ret -> right = right;
    ret -> parent = parent;
    
    return ret;
}

/*** 
 *** Old functions (from Assignment 2), largely unchanged, i.e.,
 *** except printtree and the issues related to whether the
 *** information is deep copied or not (it is now, it was not
 *** initially) :
 ***/

btree::node* btree::mknode(char* info, btree::node* l, btree::node* r, btree::node* p) const {
    char* new_info = new char [strlen(info) + 1];  // DEEP copy!!
    new_info = strcpy(new_info, info);
    btree::node* ret = new btree::node;
    ret -> info = new_info;
    ret -> parent = p;
    ret -> left = l;
    ret -> right = r;
    return ret;
}

btree::node* btree::deltree (btree::node* tree, btree::node* node) const {
    if (tree == empty) // nothing to do, cool
        return empty;
    // delete the children of our node (recursive calls):
    if (node -> left != empty) {
        deltree (tree,node -> left);
    }
    if (node -> right != empty) {
        deltree (tree,node -> right);
    }
    // node -> left, node -> right are now disposed of,
    // we set the links in the parent:
    if (node -> parent != empty) {
        if (node -> parent -> left != empty && 
            node -> parent -> left == node)
            node -> parent -> left = empty;
        if (node -> parent -> right != empty && 
            node -> parent -> right == node)
            node -> parent -> right = empty;
    }
    // finally, delete the node and return the (new) tree:
    delete node -> info;  // DEEP copy, DEEP delete!
    delete node;
    if (node == tree)
        return empty;
    else
        return tree;
}

char* btree::key (const btree::node* t) const {
    return t -> info;
}

btree::node* btree::lefttree (const btree::node* t) const {
    return t -> left;
}
btree::node* btree::righttree (const btree::node* t) const {
    return t -> right;
}

btree::node* btree::parent(const btree::node* t) const {
    return t -> parent;
}

void btree::indent(int level, ostream& out) const { // auxiliary for printtree
    for (int i = 0; i < level; i++)
        out << "   ";
    out.flush();
}

void btree::printtree (btree::node* t, ostream& out, int level) const {
    indent(level,out);
    out << "[" << t << "] ";
    if (t != empty) {
        out << "\"" << key(t) << "\"" << " ";
        out << "[parent: " << parent(t) << "]\n";
        if (lefttree(t) != empty) {
            indent(level,out);
            out << "left:\n";
            printtree(lefttree(t), out, level+1);
        }
        if (righttree(t) != empty) {
            indent(level,out);
            out << "right:\n";
            printtree(righttree(t), out, level+1);
        }
    }
    else
        out << "\n";
}
