/*
 * Part of the solution for Assignment 3 (CSC 218, Winter 2003), by
 * Stefan Bruda.
 */
#ifndef __BTREE_H
#define __BTREE_H

#include <iostream> 
using namespace std;

/*
 * The empty tree
 */
#define empty 0


/***
 *** Class for binary trees:
 ***/

class btree {

 public:

    /* 
     * Nodes in a binary tree
     */
    class node {
        char* info;
        node* left;
        node* right;
        /*
         * A node also contains its parent (empty for the root node).
         */
        node* parent;
    
        /*
         * Only these friends will be able to manipulate the information
         * therein.
         */
        friend class btree;
        friend class stree;
    
        /*
         * For easy memory trace you can use a member variable like this:
         *
         *    char dummy [10240];
         *
         * This way you can easily see when one of your structures does not
         * get deallocated. :-)
         */
    };

 protected:

    /*
     * The content of our tree.
     */
    node* the_tree;
    
    /* 
     * Helper functions, from the previous assignment (now protected).
     * There are some minor changes, see btree.cc.
     */
    node* mknode(char*, node* = empty, node* = empty, node* = empty) const;
    char* key (const node*) const;
    node* lefttree (const node*) const;
    node* righttree (const node*) const;
    node* parent(const node*) const;
    
 private:
    /*
     * Helper functions, from the previous assignment (now private).
     * There are some minor changes, see btree.cc.
     */
    node* deltree (node*, node*) const;
    void printtree (node*, ostream&, int = 0) const; // slightly changed
    void indent(int, ostream&) const;                      // auxiliary for printtree
    
    /*
     * New helper for the copy constructor and assignment operator.
     * Does the effective deep copying.
     */
    node* clone(node*, node*) const;

    /*
     * New helper for the print functions.  traverse_inorder(t,o)
     * traverses the tree rooted at t in inorder and prints to o
     * whatever it finds in there.
     */
    void traverse_inorder(const node*, ostream& = cout) const;
    
 public:
    /** Now for the member functions required in the assignment: **/

    /*
     * The usual bunch of constructors (including the copy constructor),
     * the destructor, and the assignment operator.
     */
    btree(void); 
    btree(const btree&);
    ~btree(void);
    const btree& operator=(const btree&);
    
    /*
     * zap(t) deletes from the current tree (the_tree) the subtree
     * rooted at t (deallocates memory too).  If t is not a subtree of
     * the current tree, leaves the current tree unchanged (but does
     * deallocate the space used by t).
     */
    void zap(node*);

    /*
     * print(o) prints the current tree to the stream o.  print(o,n)
     * with n !=0 prints only the content of the current tree, which is
     * traversed in inorder for this purpose.
     */
    void print(ostream& = cout, int = 0) const;
};

/*
 * The << operator: o << t is just a convenient wrapper for
 * t.print(o).
 */
ostream& operator<< (ostream&, const btree&);

#endif /* __BTREE_H */
