/*
 * Part of the solution for Assignment 3 (CSC 218, Winter 2003), by
 * Stefan Bruda.
 */
#include "stree.h"

/***
 *** Public member functions:
 ***/

// Wrapper for memtree
stree::node* stree::member(char* info) const {
    return memtree(info, the_tree);
}

// Wrapper for instree
void stree::insert(char* info) {
    the_tree = instree(info, the_tree);
}

// Wrapper for delnode
void stree::del(stree::node* node) {
    the_tree = delnode(the_tree, node);
}


/***
 *** Old functions (from Assignment 2), unchanged except delnode which
 *** does now a deep delete, and a couple of bugfixes in delnode
 *** (which are marked proeminently in the source code):
 ***/

stree::node* stree::memtree(char* info, stree::node* tree) const {
    stree::node* t = tree;
    while (1) {
        if (t == empty)
            return empty;
        else if ( strcmp(info,key(t)) == 0 )
            return t;
        else if ( strcmp(info,key(t)) < 0 )
            t = lefttree(t);
        else // strcmp(info,key(t)) > 0
            t = righttree(t);
    }
}

stree::node* stree::instree(char* info, stree::node* tree) const {
    stree::node* t = tree;
    stree::node* a = mknode(info);
    if (tree == empty)
        return a;
    while (1) {
        if ( strcmp(info,key(t)) <= 0 ) { // insert to the left
            if (lefttree(t) == empty) { // insert here, yay!
                a -> parent = t;
                t -> left = a;
                return tree;
            }
            else // not yet a leaf, go down
                t = lefttree(t);
        }
        else { // strcmp(info,key(t)) > 0, i.e., insert to the right
            if (righttree(t) == empty) { // insert here, yay!
                a -> parent = t;
                t -> right = a;
                return tree;
            }
            else // not yet a leaf, go down
                t = righttree(t);
        }
    }
}

stree::node* stree::delnode(stree::node* tree, stree::node* node) const {
    stree::node* t = node;
    // Note: We iterate the following while loop at most twice (see the
    // algorithm description in the supplementary handout).
    while (1) {
        stree::node* p = empty;
        if (t != empty)
            p = parent(t);
        if (lefttree(t) == empty) {  // no left tree, delete immediately
            // alter the link in the parent node:
            if (p != empty && p -> left == t)
                p -> left = righttree(t);
            else if (p != empty && p -> right == t)
                p -> right = righttree(t);
            else if (p == empty) { // no parent node = root
                tree = righttree(t);
                // ****************** BUG FIX ******************
                // (no longer necessary!)
                // ************** END BUG FIX ******************
            }
            // ****************** BUG FIX ******************
            // parent was not set correctly in the new child!
            if (righttree(t))
                righttree(t) -> parent = p;
            // ************** END BUG FIX ******************

            // delete the node and return:
            delete[] key(t);  // DEEP delete
            delete t;
            return tree;
        }
        else if (righttree(t) == empty) { // no right tree, delete
                                          // immediately
            // alter the link in the parent node:
            if (p != empty && p -> left == t)
                p -> left = lefttree(t);
            else if (p != empty && p -> right == t)
                p -> right = lefttree(t);
            else if (p == empty) { // no parent node = root
                tree = lefttree(t);
                // ****************** BUG FIX ******************
                // (no longer necessary!)
                // ************** END BUG FIX ******************
            }
            // ****************** BUG FIX ******************
            // parent was not set correctly in the new child!
            if (lefttree(t))
                lefttree(t) -> parent = p;
            // ************** END BUG FIX ******************
            
            // delete the node and return:
            delete[] key(t);  // DEEP delete
            delete t;
            return tree;
        }
        else { // neither righttree(t), nor lefttree(t) empty, look for
               // the maximum value in the left child, and exchange.
            stree::node* max = lefttree(t);
            // Maximum value is obtained by following the right links as
            // long as we can:
            while ( righttree(max) != empty )
                max = righttree(max);
            // ****************** BUG FIX (sort of) ******************
            // EXCHANGE the values from max and t.  NOTE: we need to
            // _exchange_ them for proper deep delete (sort of a bug fix,
            // the necessity of deep delete is only for this new program)
            char* temp = key(t);
            t -> info = key(max);
            max -> info = temp;
            // ************** END BUG FIX (sort of) ******************
            // We are now ready to delete max.  For this we set t to max and
            // then we just continue with the next iteration, which will
            // effectively delete t (which will be set to max, as desired).
            t = max;
        }
    }
}
